package ch.epfl.ppo.collections;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Random;

import org.junit.Test;

public abstract class ListTest {
    abstract public <E> List<E> newList();

    @Test
    public void isInitiallyEmpty() {
        assertTrue(newList().isEmpty());
    }

    @Test
    public void sizeIsInitially0() {
        assertEquals(0, newList().size());
    }

    @Test
    public void addIncreasesSize() {
        List<String> l = newList();
        for (int i = 0; i < 100; ++i) {
            assertEquals(i, l.size());
            l.add("bonjour");
        }
    }

    @Test(expected = IndexOutOfBoundsException.class)
    public void getFailsOnEmptyList() {
        newList().get(0);
    }

    @Test
    public void addThenGet() {
        List<Integer> l = newList();
        Random rng = new Random();
        for (int i = 0; i < 100; ++i) {
            assertEquals(i, l.size());
            int n = rng.nextInt();
            l.add(n);
            assertEquals(n, l.get(i).intValue());
        }
    }

    @Test
    public void addPreservesOrder() {
        long seed = (long)(Math.random() * 1e9);
        Random rng1 = new Random(seed);
        List<Integer> l = newList();
        for (int i = 0; i < 100; ++i)
            l.add(rng1.nextInt());
        Random rng2 = new Random(seed);
        for (int i = 0; i < 100; ++i) {
            assertEquals(rng2.nextInt(), l.get(0).intValue());
            l.remove(0);
        }
        assertTrue(l.isEmpty());
    }

    @Test
    public void removeDecreasesSize() {
        List<Integer> l = newList();
        for (int i = 0; i < 100; ++i) {
            l.add(i);
        }
        for (int i = 100; i > 0; --i) {
            assertEquals(i, l.size());
            l.remove(0);
        }
    }

    @Test
    public void removeRemovesProperElement() {
        List<Integer> l = newList();
        for (int i = 0; i < 100; ++i)
            l.add(i);
        // Remove all odd elements
        for (int i = 1; i <= 50; ++i)
            l.remove(i);
        assertEquals(50, l.size());
        for (int i = 0; i < 50; ++i)
            assertEquals(2 * i, l.get(i).intValue());
    }

    @Test(expected = IndexOutOfBoundsException.class)
    public void removeFailsWithIncorrectIndex() {
        List<Integer> l = newList();
        l.add(1);
        l.remove(1);
    }

    @Test
    public void setChangesValue() {
        List<Integer> l = newList();
        for (int i = 0; i < 100; ++i)
            l.add(i);
        for (int i = 0; i < 100; ++i)
            l.set(i, l.get(i) * 2);
        for (int i = 0; i < 100; ++i)
            assertEquals(i * 2, l.get(i).intValue());
    }

    @Test(expected = IndexOutOfBoundsException.class)
    public void setFailsWithIncorrectIndex() {
        List<Integer> l = newList();
        l.add(1);
        l.set(1, 45);
    }

    @Test
    public void iteratorIteratesInOrder() {
        List<Integer> l = newList();
        for (int i = 0; i < 100; ++i)
            l.add(i);
        Iterator<Integer> it = l.iterator();
        for (int i = 0; i < 100; ++i) {
            assertTrue(it.hasNext());
            assertEquals(i, it.next().intValue());
        }
        assertFalse(it.hasNext());
    }

    @Test(expected = NoSuchElementException.class)
    public void iteratorNextFailsIfNotHasNext() {
        this.<Integer>newList().iterator().next();
    }

    @Test(expected = IllegalStateException.class)
    public void iteratorRemoveFailsBeforeFirstNext() {
        this.<Integer>newList().iterator().remove();
    }

    @Test(expected = IllegalStateException.class)
    public void iteratorRemoveFailsWhenCalledTwice() {
        List<Integer> l = newList();
        l.add(1);
        l.add(2);
        Iterator<Integer> it = l.iterator();
        it.next();
        it.remove();
        it.remove();
    }

    @Test
    public void toStringWorksOnEmptyList() {
        assertEquals("[]", this.<Integer>newList().toString());
    }

    @Test
    public void toStringWorksOnNonEmptyList() {
        List<Integer> l = newList();
        l.add(1);
        l.add(2);
        assertEquals("[1,2]", l.toString());
    }
}
